/* AD5522 Functions, starting with register and bit definitions
 * The AD5522 is a 4 channel device. So there are 4 sets of registers, dace, calibrations, etc. 
 * There is one System Control register (SCR) accessed by clearing all 4 Channel bits. 
 * All registers are 32 bits. Actually 29 LSBs: 
 28| 27  26  25  24| 23  22  21  20| 19  18  17  16| 15  14  13  12 |11  10  09  08 |07  06  05  04|03 02 01 00|
-------------------------------------------------------------------------------------------------------------------------
 RD|CH3 CH2 CH1 CH0| M1  M0  A5  A4| A3  A2  A1  A0|                     16 bit data                           | DAC Regs

 RD| 0   0   0   0 | 0   0  CL3 CL2|CL1 CL0 CH3 CH2|CH1 CH0 CPB  DG |GAL CAL 10K GEN|G1  G0 TEN T1 |T0 LAT - - | SCL Syst                                                                               | Sys Control SCR
         
 RD|CH0 CH1 CH2 CH3| 0   0  CEN  F0| F1  -  C2  C1 | C0  ME1 ME0 FIN|SF1 SF0  CL CPO| CPV CLR -  - | -  -  -  -| PMU Ctl

AD5522 functions: 
* AD5522Init() Default settings, FV, output ON, 0V 
  AD5522Reset() HW reset
* AD5522SetRange(chan, irange)
  AD5522SetMode(chan, FVMI/FIMV)
  AD5522Sys(chan, sys)  Connect to SYS force and sense pins
  AD5522SetDac(chan, range I/V/CL/CO, x/m/c, value)
  AD5522SetVolt(chan, value)
  AD5522OnOff(chan, ON/OFF)
  
  AD5522SetMeasure (chan, I-V)
  ADCGetMeasure()

  Calibrations: Offset and Gain for each
    Force VF, IF[5]
    Clamps: VCL+/-, ICL+/-[5]
    Meas: VM, IM[5]

 */

#include "AD5522.h"

/* Get measure for one channel, with or without calibration. 
  Need to setMeasure(ch, source) first
  Range is current range for the channel. 
  cal = 0: no cal, cal = 1: cal
  chan doesn't set the AD5522 channel, it's for the data structures
  uses MMode;    // 0: MI 1: MV: 2: MT  
  data is returned, also put in Chan[ch].measV or .measI
*/
float ad5522Meas( unsigned char chan, unsigned char cal){  
  float meas = 0.0; 
  long adc, adc2;
  long ave = 64;
  unsigned char range;
  float iScale = 0.002; 
  int i;

  adc = 0;

  for (i = 0; i <ave; i++){
    adc2 = ADCGetMeasure(); 
    adc += adc2;
    //Serial.println(adc2);
  }
  meas = adc /(float)ave;

  if (Chan[chan].MMode == MV){
    meas = 10.0 * (meas - VM_C_O) / (VM_C_G); //Coarse offset and gain
    meas = (meas - Mcal[chan][VF].O) / Mcal[chan][VF].G;     // Fine calibration
    Chan[chan].measV = meas;
//  Serial.print(Chan[ch].measV, 6); Serial.println(" V");
  }
  if (Chan[chan].MMode == MI){
    //Serial.print("ADC V = "); Serial.println(meas); 
    range = Chan[ch].IRange;
    if (range ==  IR5UA) iScale = 0.000005;
    if (range ==  IR20UA) iScale = 0.00002;
    if (range ==  IR200UA) iScale = 0.0002;
    if (range ==  IR2MA) iScale = 0.002;
    if (range ==  IR50MA) iScale = 0.05;

    meas = iScale * (meas - IM_C_O) / (IM_C_G); // Scale and coarse O&G
  //  meas = (meas - Mcal[chan][IR2MA].O) / Mcal[chan][IR2MA].G;  // Fine O&G, only at 2mA
  //  meas = meas * iScale / 0.002;
    Chan[ch].measI = meas;
    //Serial.print(1000.0 * Chan[ch].measI, 8); Serial.println(" mA");
  }

  return meas;
}

//  #define PROTO    // If proto board

/* ADCGetMeasure. Read AD7685 ADC on eval board. 
  Just once, returns a short integer. 
  Averaging, scaling, cal are all above this. . 
  Since the Eval board has SDI = HI, 17 clocks are required to get all 16 bits. 
  The first 15 bits are shifted up 1, then the 16th bit added. 
  For proto, ADC SDI is wired HI
  For PCB, assert SDI High (pin 29) for conversion, output data is 16 bits. 
*/
unsigned short ADCGetMeasure(void){
  unsigned short spiRData;
  unsigned char lsb;
 #ifdef PROTO
  noInterrupts();   // critical, time-sensitive code here
  digitalWrite(ADC_CNV, HIGH);
  digitalWrite(ADC_CNV, LOW); 
  spiRData = SPI.transfer16( 0 );     // Get 15b data from ADC
  lsb = SPI.transfer( 0 );     // Get lsb, 17 clocks needed
  interrupts();
  lsb = (lsb >> 7) & 0x01;
  spiRData = (spiRData << 1) | lsb; 
  return spiRData ;
#else // New PCB
  noInterrupts();   // critical, time-sensitive code here
  digitalWrite(29, HIGH);
  digitalWrite(ADC_CNV, HIGH);
  spiRData = SPI.transfer16( 0 );     // Get 15b data from ADC
  interrupts();
  digitalWrite(ADC_CNV, LOW); 
  return spiRData ;
#endif
}

/* Set up channels for measure. Separate from ADCGetMeasure(). 
  Set all channels to HiZ except selected channel
  Measure source bits [14:13] 
  Select channel and source: 00-Isense, 01-Vsense, 10-Temp 11- HiZ
  Could store previous channel and only clear 1, not 3. 
  Make sure all 4 measure jumpers are installed.
  This function takes about 120uS to write the 4 PMUs
  Delay at end needed to let AD5522 output to change (10-20us) and to settle
  before ADC is read. 
*/
void setMeasure(int ch, int source){
  const unsigned long sourceBits = 0x03L << 13;
  ch &= 0x3;                    // Make 2 bits
  source &= 0x3L;                 // Make 2 bits

  Chan[ch].MMode = source;    // Set Measure Mode in Structs
  Instr.MChan = ch;
  
  PMU[ch] &= ~sourceBits;       // Clear old measure source bits [14:13] 
  PMU[ch] |= source << 13;      // Set Source to selected channels
  for (int i = 0; i <= 3; i++){   // Turn off all other channels
    if(i != ch){
      PMU[i] |= sourceBits;       // 11b is HiZ, so just OR with mask
      ad5522W(PMU[i]);
      }
  }
  ad5522W(PMU[ch]);
  delayMicroseconds(10);    // About 10 us delay before ADC can be read
}

void AD5522Init(){
  unsigned long PMUInit;
  unsigned long DACData; 
  int range;

  SCR = 0x003FE4A0L;              //Init SCR data from AD5522 Panel
  ad5522W(SCR);

  PMUInit = 0x00219300L;          // Ch EN, FV, MI, 2mA
  PMU[0] = AD55_CHA | PMUInit;    // Init all channels
  ad5522W(PMU[0]);
  
  PMU[0] = AD55_CH0 | PMUInit;    // Set up all PMU shadows
  PMU[1] = AD55_CH1 | PMUInit;
  PMU[2] = AD55_CH2 | PMUInit;
  PMU[3] = AD55_CH3 | PMUInit; 

  DACData = 0x0FCD8000L;          // Set all DAC voltages to 0
  ad5522W(DACData); 

  // Init the structs
  for(int i=0; i<= 3; i++){
    Chan[i].FMode = 0;
    Chan[i].IRange = IR2MA;   // I range 2mA;
    Chan[i].MMode = MI;
  }

  //setMeasure(0,0);                // Init ch0 measure, others HiZ

  //Serial.println( DACData, HEX );
  //Serial.println( spiRData, HEX );
  calC = 0x8000;      // Initial CAL values. HW Reset sets these too. 
  calM = 0xFFFF;
  
  Chan[0].forceV = 10.0;    // Initial Force V values
  Chan[1].forceV = 5.0;
  Chan[2].forceV = -5.0;
  Chan[3].forceV = -10.0;

  for( ch = 0; ch < NUM_CH; ch++){
    Chan[ch].FMode = FV;        // Set up channel structs

    // Init all Force Cal Values
    for( range = 0; range < F_CAL_NUM; range++ ){
      Fcal[ch][range].C = calC;   
      Fcal[ch][range].M = calM;  
    }

  }
  
  // // Init Measure cal values VM
  // Mcal[0][VF].O = -0.015;        // Fine voltage calibration
  // Mcal[0][VF].G = 1.0004;
  // Mcal[1][VF].O = 0.005;        // Fine calibration
  // Mcal[1][VF].G = 1.0000;
  // Mcal[2][VF].O = 0.003;        // Fine calibration
  // Mcal[2][VF].G = 1.0000;
  // Mcal[3][VF].O = 0.007;        // Fine calibration
  // Mcal[3][VF].G = 1.0000;     
  
  Mcal[0][VF].O = -0.015;        // Fine voltage calibration
  Mcal[0][VF].G = 1.0004;
  Mcal[1][VF].O = -0.010;        // Fine calibration
  Mcal[1][VF].G = 1.0011;
  Mcal[2][VF].O = -0.012;        // Fine calibration
  Mcal[2][VF].G = 1.0009;
  Mcal[3][VF].O = -0.008;        // Fine calibration
  Mcal[3][VF].G = 1.0011;


  ch = 0;
}

/* Sets FI DACs, no calibration */
void AD5522SetCurr(int ch, float value){
  unsigned long  DACData = 0;   
  long dacVal;              // Integer DAC
  unsigned long channel;
  unsigned int IOS = 85;      // Crude current offset and gain
  float IGA = 0.99641; 
  float rangeScale = 0.002;
  
  unsigned char iRange = Chan[ch].IRange; 

  if(iRange == IR5UA){
    rangeScale = 0.000005;
    DACData = DAC_FI0;
  }
  if(iRange == IR20UA){
    rangeScale = 0.000020;
    DACData = DAC_FI1;
  }
  if(iRange == IR200UA){
    rangeScale = 0.000200;
    DACData = DAC_FI2;
  }
  if(iRange == IR2MA){
    rangeScale = 0.002;
    DACData = DAC_FI3;
  }
  if(iRange == IR50MA){  
    rangeScale = 0.025;
    DACData = DAC_FI4;
  }
  
  dacVal = (value * 29127.0 * IGA / rangeScale) + 32768.0 - IOS; 
  // dacVal limits
  if(dacVal > 0xFFFF) dacVal = 0xFFFF;
  if(dacVal < 0) dacVal = 0;
  //Simple test for chan limits. Better to use case statement?
  if((ch >= 0) && (ch <= 4 )) {
    if(ch == 0) channel = AD55_CH0;
    if(ch == 1) channel = AD55_CH1;
    if(ch == 2) channel = AD55_CH2;
    if(ch == 3) channel = AD55_CH3;
    if(ch == 4) channel = AD55_CHA;
    
    // Range is set above. 
    DACData |= channel | AD55_M_X | dacVal ; 
    ad5522W(DACData);  
  }
}

/* UNUSED Currently only sets voltage DAC, no calibration */
void AD5522SetVolt(int ch, float value){
  unsigned long dacVal, DACData;
  unsigned long channel;
//  dacVal = (value * 29127.0 / 10.0) + 32768;  
  dacVal = (value * 29140.0 / 10.0) + 32768;  

  //Simple test for chan limits. Better to use case statement?
  if((ch >= 0) && (ch <= 4 )) {
    if(ch == 0) channel = AD55_CH0;
    if(ch == 1) channel = AD55_CH1;
    if(ch == 2) channel = AD55_CH2;
    if(ch == 3) channel = AD55_CH3;
    if(ch == 4) channel = AD55_CHA;
    DACData = channel | AD55_M_X | DAC_FV0 | dacVal ; 
    ad5522W(DACData); 
  }
}

/* Sets channel offset and gain with Cal factors 
  No need to set cal in final code. It is handled elsewhere
*/
void AD5522SetVoltCal(int ch, float value ){
  unsigned long dacVal, DACData;
  unsigned long channel;
// dacVal = (value * 29127.0 / 10.0) + 32768.5;  
  dacVal = (value * 29140.0 / 10.0) + 32768;  
  //Simple test for chan limits. Better to use case statement?
  if((ch >= 0) && (ch <= 4 )) {
    if(ch == 0) channel = AD55_CH0;
    if(ch == 1) channel = AD55_CH1;
    if(ch == 2) channel = AD55_CH2;
    if(ch == 3) channel = AD55_CH3;
    if(ch == 4) channel = AD55_CHA;

    DACData = channel | AD55_M_C | DAC_FV0  | Fcal[ch][VF].C ;  // Offset
    ad5522W(DACData); 
    DACData = channel | AD55_M_M | DAC_FV0  | Fcal[ch][VF].M ;  //  Gain
    ad5522W(DACData); 
    DACData = channel | AD55_M_X | DAC_FV0 | dacVal ;  // Set X DAC last, after M and C
    ad5522W(DACData); 
    // Need to write values to EEPROM
  }
}

// Set Force mode: FV, FI, FHIZ. PMU bits[20:19]
void AD5522SetFMode(unsigned char chan, unsigned long fMode){
  const unsigned long fModeBits = 0x03L << 19;

  chan &= 0x3;                  // Make range 2 bits
  fMode &= 0x3L;                // Make F 2 bits   
  Chan[chan].FMode = fMode;     // Set structure 
  PMU[chan] &= ~fModeBits;      // Clear old Force mode bits
  PMU[chan] |= fMode << 19;     // Set new mode
  ad5522W(PMU[chan]);  
}

// Set irange, PMU bits [17:15]
void AD5522SetIRange(unsigned char chan, unsigned char iRange){
  const unsigned long rangeBits = 0x07L << 15;
  if (iRange >= 0 && iRange <= 4){  // Check for valid
    chan &= 0x3;                  // Make chan 2 bits
 //   if (iRange > 4) iRange = 4;
 //   iRange &= 0x7L;               // Make irange 3 bits   
    Chan[chan].IRange = iRange;   // Set Struct
    PMU[chan] &= ~rangeBits;      // Clear old I range bits [17:15] 
    PMU[chan] |= iRange << 15;    // Set new range
    ad5522W(PMU[chan]);
  }
  else Serial.println("IR out of range");
}


void printPMU(void){    // Display all PMUs
  for (int i = 0; i <= 3; i++){    
    Serial.print( "PMU [" ); Serial.print( i ); Serial.print( "] = 0x") ;
    Serial.println( PMU[i], HEX );
  }
}

void printCal(int ch){

  Serial.print ("Ch"); Serial.print (ch);
  Serial.println(" IF0  IF1  IF2  IF3  IF4  VF   CLLI CLHI CLLV CLHV");
  Serial.println("-----------------------------------------------------");
  Serial.print ("C:  "); 
  for (int i= 0; i < F_CAL_NUM; i++){
    Serial.print( Fcal[ch][i].C, HEX);
    Serial.print(" ");
  }
  Serial.println();
  
  Serial.print ("M:  ");
  for (int i= 0; i < F_CAL_NUM; i++){
    Serial.print( Fcal[ch][i].M, HEX);
    Serial.print(" ");
  }
  Serial.println();
}

/* Low Level SPI */

/* Read data from AD5522 SCR, PMU or DAC registers
Chan is 0-3 for PMU, 4 for SCR. 
Write 32 bits, then read 24 bits. 
24Bit positions are the same as Write[23:0] */
unsigned long ad5522RD(int ch, int reg ){
  unsigned long spiRData = 0L, spiWData = 0L, DACReg, channel, mode;
  
  SPI.beginTransaction(SPISettings(4000000, MSBFIRST, SPI_MODE1));

  mode = (reg & 0x03) << 22;        // 0: Register, 1:M, 2: C, 3 X1
  
  if(ch == 0) channel = AD55_CH0;   // PMU Register
  if(ch == 1) channel = AD55_CH1;
  if(ch == 2) channel = AD55_CH2;
  if(ch == 3) channel = AD55_CH3;
  if(ch == 4) channel = 0L;         // SCR
  
  if (mode == 0) DACReg = 0;       
  if (mode == 1) DACReg = DAC_FV0;
  if (mode == 2) DACReg = DAC_FV0;
  if (mode == 3) DACReg = DAC_FV0;
  
  spiWData = (1L << 28) | channel | mode  | DACReg ;    // Build read command

  digitalWrite(AD5522_CS, LOW);
  SPI.transfer16( spiWData >> 16) ;         // Send MSBs
  SPI.transfer16( spiWData & 0x0000FFFFL ) & 0x0000FFFFL ; // LSBs
  digitalWrite(AD5522_CS, HIGH); 
  digitalWrite(AD5522_CS, LOW);
  spiRData = SPI.transfer16( 0xFFFF) << 8;      // Get MSBs Send NOP, 0xFFFFFF
  spiRData |= SPI.transfer( 0xFF) & 0x000000FFL ; // 8 LSBs. Send NOP 
  digitalWrite(AD5522_CS, HIGH);

  //SPI.beginTransaction(SPISettings(4000000, MSBFIRST, SPI_MODE0));
  Serial.print( "Read Cmd = 0x");
  Serial.print( spiWData, HEX );
  Serial.print( " Read Data = 0x");
  Serial.println( spiRData, HEX );
  return spiRData;

}

// Write 32 bit SPI data to AD5522. Two 16b transfers
void ad5522W(unsigned long spiWData){
  SPI.beginTransaction(SPISettings(4000000, MSBFIRST, SPI_MODE1));
  digitalWrite(AD5522_CS, LOW);
  SPI.transfer16(spiWData >> 16); // MSBs
  SPI.transfer16(spiWData & 0x0000FFFFL) ; // LSBs
  digitalWrite(AD5522_CS, HIGH);
  //SPI.beginTransaction(SPISettings(4000000, MSBFIRST, SPI_MODE0));
}

